program s
  use input_mdl, only: muUV,muIR,muMS,dp
  use modelinput, only: linc,printp,rs,scalefac
  use pde_integrator, only: rk4_step, rk45_step,euler_step, update_func, set_kernel
  use class_particles, only: print_parameters_mdl
  implicit none

  character(len=10) :: renoscheme = ''
  complex(dp) :: t,Gf,minv,minmh2,minmsb2
  integer :: i
  real(dp) :: sa_in,l3_in,mhh_in,massreg

  integer, parameter :: steps = 400
  real(dp), parameter :: scale_init = 200d0
  real(dp), dimension(-steps-1:steps+1) :: scale_arr, sa_arr, l3_arr

  procedure (update_func), pointer :: reset_func
  procedure (set_kernel), pointer  :: set_k

  character(len=6) :: benchmarkpoint = 'run'
  character(len=100) :: file_bp
  character(len=*), parameter :: fmt = '(e21.14,x,e21.14,x,e21.14)'
  complex(dp),parameter :: cone=1d0,cnul=0d0

  reset_func => reset_alpha_l3_hs
  set_k => set_beta_alpha_l3_hs

  call read_input

  renoscheme = trim(rs)

  if (printp .gt. 0) call print_parameters_mdl(printp-1)

  call reset_alpha_l3_hs
  call running(scale_init,scale_init/scalefac)
  call reset_hs_param
  call reset_alpha_l3_hs
  call running(scale_init,scale_init*scalefac)

  file_bp = trim(benchmarkpoint) // '_' // trim(renoscheme)
  open(unit=1,file=trim(file_bp),status='replace',position='append')

  write(1,'(a,14x,a,16x,a)') '# mu/mu0', 'sa(mu)', 'l3(mu)'
  do i = -steps, steps
    if (i .ne. 0) then
      write(1,fmt) scale_arr(i), sa_arr(i), l3_arr(i)
    else
      write(1,fmt) 1d0, sa_in, l3_in
    end if
  end do

  close(1)
  close(2)

  contains  

  subroutine read_input
    use fill_couplings, only: fill_couplings_mdl
    use hs_ct, only: initct
    use collier_interface, only: init_collier
    use modelinput, only: read_modelinput,sa,l3,MHH,printp,scalefac
    use modelfile, only: set_parameter_mdl
    character(len=*), parameter :: fmt0 = '(a,f21.15)'
    character(len=*), parameter :: fmt = '(2x,a,x,f23.14)'
    character(len=*), parameter :: fmt_a = '(2x,a,x,a)'

    call read_modelinput

    muUV = 100d0
    muIR = 100d0
    muMS = 100d0

    mhh_in = MHH
    sa_in = sa
    l3_in = l3

    if (printp .eq. 0) then
      write(*,fmt) "mhh in:", mhh_in
      write(*,fmt) "sa  in:", sa_in
      write(*,fmt) "l3  in:", l3_in
      write(*,fmt_a) "rs:", trim(rs)
    end if

    sa_arr(0) = sa_in
    l3_arr(0) = l3_in
    scale_arr(0) = 1d0

    call reset_hs_param

    muMS = (125.09d0 + mhh_in)/2d0
    muUV = muMS
    call set_parameter_mdl('muMS_BSM', muMS*cone)

    write(*,fmt0) " sa(mu0)", sa_in
    write(*,fmt0) " l3(mu0)", l3_in

    call init_collier

  end subroutine read_input

  subroutine reset_hs_param
    use modelfile, only: set_parameter_mdl
    call set_parameter_mdl('sa', sa_in*cone)
    call set_parameter_mdl('l3', l3_in*cone)
    call set_parameter_mdl('MHH', mhh_in*cone)
    call set_parameter_mdl("WHH",cnul)
  end subroutine reset_hs_param

  subroutine running(scale0,scale)
    use class_particles, only: sa,l3,MHl,MHh,vev,l2
    use pde_integrator, only: rk4_step, euler_step
    real(dp), intent(in)    :: scale0,scale
    real(dp) :: a0,l30,am,l3m
    complex(dp),dimension(2) :: y
    real(dp) :: scale20,scale2,stepsize,scale2m,stepsize_tmp,sar
    integer :: step
    character(len=30) :: fmt = '(a,f4.2,a,f21.15)'

    scale20 = scale0**2d0
    scale2  = scale**2d0
    step = 0

    stepsize = (scale2-scale20)/real(steps,kind=kind(23d0))

    l30 = l3
    sar=sa
    a0 = asin(sar)

    call reset_alpha_l3_hs

    scale2m = scale20
    am      = a0
    l3m     = l30
    do while ((scale2m .gt. scale2 .and. stepsize .lt. 0d0) .or. &
              (scale2m .lt. scale2 .and. stepsize .gt. 0d0))

      y = (/am, l3m/)
      !call euler_step(y,scale2m,stepsize,reset_func,set_k)
      call rk4_step(y,scale2m,stepsize,2,reset_func,set_k)
      !call rk45_step(y,scale2m,stepsize,2,reset_func,set_k)
      am = y(1)
      l3m = y(2)

      sa = sin(am)
      l3 = l3m

      call reset_alpha_l3_hs
      if (stepsize .gt. 0) then
        step = step + 1
      else
        step = step - 1
      end if
      l3_arr(step) = real(l3,kind=kind(23d0))
      sa_arr(step) = sin(am)
      scale_arr(step) = sqrt(scale2m)/scale_init
    end do
    sa = sin(am)
    l3 = l3m
    
    call reset_alpha_l3_hs

    if (scale .lt. scale0) then
      write(*,fmt) " sa(mu0/", real(scale0/scale), ")", sin(am)
      write(*,fmt) " l3(mu0/", real(scale0/scale), ")", real(l3,kind=kind(23d0))
    else
      write(*,fmt) " sa(", real(scale/scale0), " mu0)", sin(am)
      write(*,fmt) " l3(", real(scale/scale0), " mu0)", real(l3,kind=kind(23d0))
    end if

  end subroutine running

  subroutine set_beta_alpha_l3_hs(x,f)
    use hs_ct, only: da_FJTS, da_PRTS, dl3_l1_MS_OS, &
                     dl3_l1_MS_FJTS, dl3_l1_MS_PRTS 
    real(dp),                  intent(in)  :: x
    complex(dp), dimension(:), intent(out) :: f

    if (renoscheme .eq. 'FJ') then
      f(:) = (/da_FJTS/x,dl3_l1_MS_FJTS/x/)
    else if (renoscheme .eq. 'PR') then
      f(:) = (/da_PRTS/x,dl3_l1_MS_PRTS/x/)
    else if (renoscheme .eq. 'OS') then
      f(:) = (/cmplx(0d0,kind=kind(23d0)),dl3_l1_MS_OS/x/)
    else
      write(*,*) "Requested RS: ",renoscheme
      write(*,*) "but not available."
      write(*,*) "Allowed schemes: FJ, PR, OS"
      stop 9
    end if

  end subroutine set_beta_alpha_l3_hs

  subroutine reset_alpha_l3_hs(yc)
    use class_particles, only: sa,l3,init_parameters_mdl
    use hs_ct, only: initct, da_FJTS, da_PRTS, &
                     dl3_l1_MS_OS, dl3_l1_MS_FJTS, dl3_l1_MS_PRTS
    complex(dp), dimension(:), optional, intent(in) :: yc
    complex(dp) :: da_PRTS_tmp
    complex(dp) :: da_l1_MS_tmp, dl3_l1_MS_OS_tmp, &
                                dl3_l1_MS_FJTS_tmp, dl3_l1_MS_PRTS_tmp

    if (present(yc)) then
      sa = sin(yc(1))
      l3 = yc(2)
    endif

    call init_parameters_mdl()
    call initct(0d0)
    da_PRTS_tmp = da_PRTS
    dl3_l1_MS_OS_tmp = dl3_l1_MS_OS
    dl3_l1_MS_FJTS_tmp = dl3_l1_MS_FJTS
    dl3_l1_MS_PRTS_tmp = dl3_l1_MS_PRTS
    call initct(1d0)
    da_PRTS = da_PRTS - da_PRTS_tmp
    dl3_l1_MS_OS = dl3_l1_MS_OS - dl3_l1_MS_OS_tmp
    dl3_l1_MS_FJTS = dl3_l1_MS_FJTS - dl3_l1_MS_FJTS_tmp
    dl3_l1_MS_PRTS = dl3_l1_MS_PRTS - dl3_l1_MS_PRTS_tmp

  end subroutine reset_alpha_l3_hs

end program s
